#==============================================================================
#
#  $Id: dopy.py,v 1.9 1999/07/09 13:48:10 mike Exp $
#
"""
   The DOPY main module.  This houses all of the central classes and base
   classes that comprise the core DOPY system.
"""
#
#  Copyright (C) 1999 Michael A. Muller
#
#  Permission is granted to use, modify and redistribute this code,
#  providing that the following conditions are met:
#
#  1) This copyright/licensing notice must remain intact.
#  2) If the code is modified and redistributed, the modifications must 
#  include documentation indicating that the code has been modified.
#  3) The author(s) of this code must be indemnified and held harmless
#  against any damage resulting from the use of this code.
#
#  This code comes with ABSOLUTELY NO WARRANTEE, not even the implied 
#  warrantee of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#  $Log: dopy.py,v $
#  Revision 1.9  1999/07/09 13:48:10  mike
#  Added hooks for dopy_beforeMethod() and dopy_afterMethod()
#
#  Revision 1.8  1999/06/16 14:32:22  mike
#  First cut at name services.
#
#  Revision 1.7  1999/06/14 17:50:15  mike
#  Added the ability to preserve tracebacks across remote method invocations.
#
#  Revision 1.6  1999/06/11 14:03:55  mike
#  Made remote objects transportable.
#
#  Revision 1.5  1999/04/24 21:51:30  mike
#  Bug fixes: context object was never getting released (causing the protocol to
#  never be released).  Switched to traceback module for displaying tracebacks
#  from mainloop CLI.
#
#  Revision 1.4  1999/04/22 00:30:19  mike
#  Added the Context facility to the hub.
#
#
#==============================================================================


"""
   Limitations
   -----------
   
   At this time, output parameters are not supported (except by passing in
   remote objects).
   
   XXX should protocol and protocol info objects be one and the same?
   Or would this only pollute the protocol interface?  There is a speed
   advantage to this - but I think that protocol info may need to manage
   the underlying connection object.
   
"""

import thread, string

import tb

class Error(Exception):
   """
      Base class for all DOPY errors.
   """
   pass

class InvalidMethod(Error):
   """
      Raised when an invalid method name is cited in a request.
   """
   pass

class InvalidObjectKey(Error):
   """
      Raised when an invalid object key is cited as the target for a 
      request.
   """
   pass

class BadResponseError(Error):
   """
      Raised when a response is received with an ID other than that expected.
   """
   pass

class ProtocolTerminatedError(Error):
   """
      Raised when a protocol is terminated.
   """
   pass

class ImplementedBySubclassError(Error):
   """
      Raised by base class methods that are expected to be implemented
      by derived classes.
   """
   pass

class _Abort(Exception):
   """
      Internal error raise when a failed precondition aborts request
      processing.
   """
   pass

class RemoteObject:
   
   """
      The proxy object.
      
      The following public variables exist:
      
      `dopy_protoSpec`::
         An instance of *Protocol* defining the protocol spec for the
         object.  This may also be *None*, indicating that the object
         exists locally.  If it is *None*, there must be a local object
         defined for it.
      `dopy_key`::
         The key for the object.  Should be unique within the protocol
         spec.
      
      Protected variables:
      
      `_dopy_localObject`::
         For a true remote object, this is set to *None*.  If 
         /dopy_protoSpec/ is *None*, this contains the object that this
         is a proxy for.         
   """
   
   def __init__(self, protoSpec, key, localObject = None):
      self.dopy_protoSpec = protoSpec
      self.dopy_key = key
      self._dopy_localObject = localObject
      if localObject:
         assert protoSpec is None

   def __getstate__(self):
      dict = self.__dict__.copy()
      if self.dopy_protoSpec:
         dict['dopy_protoSpec'] = self.dopy_protoSpec.getProtocolId()
      
      # local objects _never_ get flattened: on the other side they must have
      # protocols.
      dict['_dopy_localObject'] = None
      
      return dict

   def __setstate__(self, dict):
      self.__dict__ = dict
      
      # if the protocol spec is not none, it is a string identifying the
      # protocol.  If it is none, the _current_ protocol is the one that
      # will be used.
      if self.dopy_protoSpec:
         self.dopy_protoSpec = _hub.getProtocol(self.dopy_protoSpec)
      else:
         self.dopy_protoSpec = _hub.getContext().protocol()
   
class RemoteMethod:
   """
      A remote method stores information about an operation that can
      be performed on a remote object.
   """
   
   def __init__(self, target, name):
      self.target = target
      self.name = name
   
   def __call__(self, *parms, **kwParms):
      req = Request(self.target, self.name, parms, kwParms)
      response = _hub.processRequest(req)
      if response.exception is not None:
         raise response.exception
      else:
         return response.returnValue

class DynamicRemoteObject(RemoteObject):   
   """
      A dynamic remote object has no facility for local validation of
      method invocations: it has no proxy.  Therefore, it is the
      easiest way to execute things remotely.  Attribute errors are very
      costly, though.
   """
   
   def __getattr__(self, attr):
      # allows one of these to be safely pickled
      if attr != '__getinitargs__':
         
         # if the object exists locally, delegate this.
         if self._dopy_localObject:
            return getattr(self._dopy_localObject, attr)
         else:      
            return RemoteMethod(self, attr)
            
      else:
         raise AttributeError()

class RemoteObjectFactory:
   """
      Derived classes create remote objects.
   """
   
   def create(self, protoSpec, key, localObject = None):
      """
         Derived classes must override this function to provide one that
         returns their particular variety of remote object.
      """
      raise ImplementedBySubclassError()

class DynamicRemoteObjectFactory(RemoteObjectFactory):
   """
      Creates dynamic remote objects.
   """
   
   def create(self, protoSpec, key, localObject = None):
      return DynamicRemoteObject(protoSpec, key, localObject)

class Request:

   """
      Contains all of the information necessary to invoke a remote method.
      Public variables:
      /id/::
         The message id which is used to associate the request with its
         response (set by the hub before the request is sent).
      /target/::
         The object key that the method is being invoked upon.
      /method/::
         The name of the method.
      /parms/::
         The parameter sequence.
      /kwParms/::
         The keyword parameters.
   """
   
   def __init__(self, target, method, parms, kwParms):
      self.id = None
      self.target = target
      self.method = method
      self.parms = parms
      self.kwParms = kwParms

   def __getstate__(self):
      # replace the object with its key so that we don't get into
      # an infinitely recursive loop looking for __getinitargs__ with 
      # __getattr__
      dict = self.__dict__.copy()
      dict['target'] = self.target.dopy_key
      return dict

class Response:
   """   
      Response to a Request.  Public variables:
      /id/::
         Response id.  This is the same id as that of the matching request.
      /returnValue/::
         The return value of the remote method.
      /exception/::
         The exception raise by the remote method (*None* if no
         exception was raised).
   """
   
   def __init__(self, id, returnValue = None, exception = None):
      self.id = id
      self.returnValue = returnValue
      self.exception = exception

class _AwaitingResponse:
   
   """
      Instances of this class track pending requests.  They are managed by
      the Hub in the sendRequest() method.
      
      Public variables:
      /id/::
         The key that this object is stored under.  The same id as is used in 
         the matching Request and Response objects.
      /response/::
         Allows us to store the response so that it can be returned to the
         caller.
      /threaded/::
         A class level variable that is used to indicate whether we are using
         the threaded model or not.
   """
   
   threaded = 1
   
   def __init__(self, id):
      """
         Constructs a new instance using the given id.
      """
      if self.threaded:
         self.__wait = thread.allocate_lock()
         self.__wait.acquire()
      self.id = id
      self.response = None
   
   def responseReceived(self, response):
      """
         Used to indicate that a response has been received and to set the
         response that was received.  Any threads waiting on getResponse()
         will continue.
      """
      self.response = response
      if self.threaded:
         self.__wait.release()
   
   def getResponse(self):
      """
         Waits for a response to be received and then returns it.
      """
      if self.threaded:
         self.__wait.acquire()
      else:
         self.response = self.hub.getResponse(self.id)
      return self.response

class FileEventSource:
   
   """
      This is any kind of object that can be used within the select() method
      as a source of low-level events.
   """
   
   def fileno(self):
      """
         Derived classes should implement this to provide a method that
         returns a file number for the object, suitable for use in the
         select() function.
      """
      raise ImplementedBySubclassError()

   def isActive(self):
      """
         Returns true for as long as the server is active.
      """
      raise ImplementedBySubclassError()
   
   def shutdown(self):
      """
         Shuts down the connection.
      """
      raise ImplementedBySubclassError()

class ProtocolServer(FileEventSource):
   
   """
      Protocol servers handle incoming connections on a particular protocol.
   """
   
   def acceptConnection(self):
      """
         Returns a new connection for the server.
      """
      raise ImplementedBySubclassError()   
      

class Protocol(FileEventSource):
      
   """
      Handles the nitty-gritty details of sending and recieving requests
      and responses across a communication protocol.
      
      Since protocols must be unpickleable, there may be multiple instances
      of the same protocol object.  Therefore, there should generally be
      some object that the protocol can locate within its __setattr__ method
      that actually houses the single instance of the connection that
      the multiple protocol objects refer to.
   """
   
   def getNext(self):
      """
         Derived classes must implement this to get the next object
         received on the protocol.  Should return *None* if the
         connection closes.
      """
      raise ImplementedBySubclassError()

   def send(self, msg):
      """
         Sends a request or response message through the protocol.
      """
      raise ImplementedBySubclassError()

   def getProtocolId(self):
      """
         Derived classes must use implement this to return a unique protocol
         id.  Protocol ids should be strings beginning with a type
         identifier followed by a colon (e.g. "tcp:") and followed by a value
         that uniquely identifies the connection.
      """
      raise ImplementedBySubclassError()      

class ProtocolInfo:

   """
      Protocol info manages the current message id and a list of requests.
      Public variables:
      /requests/::
         A dictionary mapping message ids to request locks.
      /proto/::
         The Protocol object itself.
   """
      
   def __init__(self, protocol):
      """
         Constructs a ProtocolInfo object from the Protocol itself.
      """
      self.__lk = thread.allocate_lock()
      self.__counter = 0
      self.requests = {}
      self.proto = protocol
   
   def getNextId(self):
      """
         Returns the next available message id.
      """
      self.__lk.acquire()
      rc = self.__counter
      self.__counter = self.__counter + 1
      self.__lk.release()
      return rc

class Context:
   """
      Instances of this class manage information about the caller context
      which may be used during the processing of a remote method.
   """

   def __init__(self, protocol):
      self.__protocol = protocol
   
   def protocol(self):
      """
         Returns the protocol which was used to communicate to the hub.
      """
      return self.__protocol

class Hub:
   
   """
      The Hub is the equivalent of the ORB - a central context for everything
      in the distributed object system.
      
      The Hub has a *registry* which can be as simple as a dictionary
      or as sophisticated as a CORBA NamingContext.  It must provide
      the __getitem__ and __setitem__ methods to dereference and assign
      keys to objects.
   """
   
   def __init__(self, registry):
      self.__registry = registry
      self.__server = None
      self.__threaded = 1
      self.__protocolInfo = {}
      if self.__threaded:
         self.__context = {}
      else:
         self.__context = None

   def isThreaded(self):
      """
         Returns true if the hub is threaded, false if not.
      """
      return self.__threaded
   
   def addServer(self, server):
      """
         Registers /server/ (an instance of ProtocolServer).  The hub will
         begin handling connections on the server.
      """
      self.__server = server
      if self.__threaded:
         thread.start_new_thread(self.__serviceConnections, (server,))

   def addObject(self, key, object):
      """
         Registers the /object/ as a remote object under the name /key/ (a 
         string).  Any kind of object can be registered.
         
         Note that a key consisting of name parts seperated by forward
         slashes can be used, but it does not produce the results that one
         might expect after reading the documentation on @getObject() - i.e.
         the object is not stored in a nested context.
      """
      self.__registry[key] = object

   def __getObject(self, node, splitKey):
      """
         Private version of @getObject(), in which /splitKey/ is a list.
      """
      node = node[splitKey[0]]
      remainder = splitKey[1:]
      if remainder:
         return self.__getObject(node, remainder)
      else:
         return node

   def getObject(self, fullKey):
      """
         Returns the object (the object that was added using #addObject()#,
         not necessarily a @RemoteObject) associated with /fullKey/.
         
         /fullKey/ can resemble a UNIX path - that is, it can consist
         of name components seperated by forward slashes.  If it is, 
         and there is nothing matching the key in the registry, 
         a recursive retrieval is performed.
      """
      splitKey = string.split(fullKey, '/')
      if not self.__registry.has_key(fullKey):
         return self.__getObject(self.__registry, splitKey)
      else:
         return self.__registry[fullKey]

   def makeAnonObject(self, object, remoteFactory = None):
      """
         Makes /object/ into an anonymous remote object.  Returns an instance 
         of a @RemoteObject derivative for the object.
         
         Anonymous remote objects are remote objects that are registered
         without any particular key - presumably for temporary use.  Actually,
         the key is the object's *id* - the address of the object.
         
         If /remoteFactory/ is supplied, it should be an instance of a class
         derived from @RemoteObjectFactory.  If it is not supplied, 
         @DynamicRemoteObjectFactory is used to create and return a 
         @DynamicRemoteObject.
      """
      key = `id(object)`
      self.addObject(key, object)
      if not remoteFactory:
         remoteFactory = DynamicRemoteObjectFactory()
      return remoteFactory.create(None, key, object)

   def makeRemoteObject(self, objectId, remoteFactory = None):
      """
         Returns a remote object for the given object id.  An object id
         consists of a protocol id followed by a colon and the object key.
         
         If /remoteFactory/ is supplied, it should be an instance of a class
         derived from @RemoteObjectFactory.  If it is not supplied, 
         @DynamicRemoteObjectFactory is used to create and return a 
         @DynamicRemoteObject.
      """
      # break the object id up into the protocol id and the object key
      parsedId = string.split(objectId, ':')
      key = parsedId[-1]
      protocolId = string.join(parsedId[:-1], ':')
      
      if not remoteFactory:
         remoteFactory = DynamicRemoteObjectFactory()
      return remoteFactory.create(self.getProtocol(protocolId), key)

   def addProtocol(self, proto):
      """
         This function should generally be called by the protocol constructor
         to register itself with the hub.
      """
      self.__protocolInfo[proto.getProtocolId()] = ProtocolInfo(proto)
      if self.__threaded:
         thread.start_new_thread(self.__serviceClient, (proto,))

   def getProtocol(self, protoId):
      """
         This is used to get or create a protocol from its id.  Returns the
         protocol if it already exists, creates it if it doesn't.
      """
      try:
         return self.__protocolInfo[protoId].proto
      except KeyError, ex:
         global protocolFactory
         return protocolFactory.create(protoId)

   def getContext(self):
      """
         Returns the current context object.  This is only available from
         within a remote method call.  If it is called from somewhere
         other than a remotely invoked method, *None* is returned.
      """
      if self.__threaded:
         return self.__context.get(thread.get_ident())
      else:
         return self.__context

   def processRequest(self, request):
      """
         Processes a request (an instance of @Request) and returns
         the response (a @Response).
      """
      # get the protocol and protocol info objects
      proto = request.target.dopy_protoSpec
      protoInfo = self.__protocolInfo[proto.getProtocolId()]
      
      # give the request a unique id
      request.id = protoInfo.getNextId()
      
      # build the object that we will use to get our response
      resp = _AwaitingResponse(request.id)
      protoInfo.requests[request.id] = resp
      
      # send the request
      proto.send(request)
      
      # get and return the response
      return resp.getResponse()
   
   def __serviceConnections(self, server):
      """
         This is executed in a thread to begin servicing client connections
         for a particular server.
      """
      while server.isActive():
         server.acceptConnection()

   def __connectionClosed(self, protoId, protoInfo):
      """
         Called when a connection is closed.
      """
      for waiter in protoInfo.requests.values():
         resp = Response(waiter.id, 
                         exception = ProtocolTerminatedError(protoId)
                         )
         waiter.responseReceived(resp)
      del self.__protocolInfo[protoId]
   
   def __serviceClient(self, clientConnection):
      myid = clientConnection.getProtocolId()
      proto = self.__protocolInfo[myid]
      
      # context will be associated with this thread.
      self.__setContext(Context(clientConnection))
      
      while clientConnection.isActive():      
         request = clientConnection.getNext()
         
         # if the getNext() method returned None, the connection has gone
         # away: clean up and exit
         if not request:
            self.__connectionClosed(myid, proto)
            break
            
         elif isinstance(request, Response):
            proto.requests[request.id].responseReceived(request)
         else:
            thread.start_new_thread(self.__serviceFunction, 
                                    (clientConnection, request)
                                    )
            
      self.__clearContext()

   def __serviceFunction(self, clientConnection, request):
      self.__setContext(Context(clientConnection))
      response = self.__processRequest(request, clientConnection)
      clientConnection.send(response)      
      self.__clearContext()
   
   def __processRequest(self, request, protocol):
      try:
         try:
            obj = self.getObject(request.target)
         except Exception, ex:
            # check for all kinds of exceptions because an attempt to resolve
            # an object could end up going into user code.
            response = Response(request.id, exception = InvalidObjectKey(ex))
            raise _Abort()
            
         try:
            meth = getattr(obj, request.method)
         except AttributeError, ex:
            response = Response(request.id, exception = InvalidMethod(ex))
            raise _Abort()
         
         try:
            # call the before function if it exists
            if hasattr(obj, 'dopy_beforeMethod'):
               obj.dopy_beforeMethod(request)
               
            # call the method and create the response message
            retval = apply(meth, request.parms, request.kwParms)
            response = Response(request.id, returnValue = retval)
                           
         except Exception, ex:         
            tb.saveTraceback(ex)
            response = Response(request.id, exception = ex)
         
         try:
         
            # call the after function if it exists
            if hasattr(obj, 'dopy_afterMethod'):
               obj.dopy_afterMethod(request, response)
         except:
            pass
            
      except _Abort, ex:
         pass
      return response      

   def __selectLoop(self, msgId):
      wr = rd = er = [self.__server]
      while 1:
         wrx, rdx, erx = select.select(wr, rd, er)
         while rdx:
            s = rdx.pop()
            if isinstance(s, ProtocolServer):
               wr.append(s.acceptConnection())
            else:
               self.__setContext(Context(s))
               req = s.getNext()
               if isinstance(req, Response):
                  if req.id != msgId:
                     raise BadResponseError("message id %s received, %s "
                                             "expected." % msgId)
                  return req
               else:
                  resp = self.__processRequest(req, s)
                  s.send(resp)            
               self.__clearContext()

   def __setContext(self, context):
      if self.__threaded:
         self.__context[thread.get_ident()] = context
      else:
         self.__context = context
   
   def __clearContext(self):
      if self.__threaded:
         del self.__context[thread.get_ident()]
      else:
         self.__context = None
   
   def mainloop(self):
      if self.__threaded:
         import traceback
         while 1:
            inp = raw_input()
            if inp[:2] == '@q':
               return
            try:
               exec inp
            except SystemExit, ex:
               raise
            except Exception, ex:
               traceback.print_exc()
      else:
         self.__selectLoop()

   def _getResponse(self, responseId):
      """
         This function reenters the message processing loop and waits for a
	 particular response.

         If any response other that response should be received, a 
         @BadResponse exception will be raised.
      """
      assert not self.__threaded
      return self.__selectLoop(responseId)

class ProtocolFactory:
   """
      The protocol factory is a singleton that is always in existence - it
      manages the creation functions for different protocols.
      
      Protocols should each provide a factory method that creates the protocol
      from a list of strings - these being the fields (as divided by colons)
      of the protocol id.
   """

   def __init__(self):
      self.protos = {}
   
   def register(self, protoTypeId, protoFact):
      self.protos[protoTypeId] = protoFact
   
   def create(self, protoId):
      name = string.split(protoId, ':')
      return self.protos[name[0]](name)

protocolFactory = ProtocolFactory()

_hub = None

def getHub():
   global _hub
   if not _hub:
      _hub = Hub({})
   return _hub
